      * This program gets the current weather from the
      * National Oceanic and Atmospheric Administration (NOAA)'s
      * National Digital Forecast Database (NDFD)
      * web service and saves it into a message file.
      *      http://graphical.weather.gov/xml/
      *                                   Scott Klement, April 24, 2008
      *
      * REQUIREMENTS:
      *   - V5R2 or later.
      *   - You must have TCP/IP installed and running.
      *   - You must be able to resolve domain names for Internet hosts.
      *        (i.e. convert www.weather.gov into an IP address.)
      *   - You must have HTTPAPI installed
      *        (see http://www.scottklement.com/httpapi )
      *   - You must put LIBHTTP in your library list.
      *
      *
      * TO COMPILE THIS:
      *> ign:   ADDLIBLE LIBHTTP *LAST
      *>        CRTBNDRPG LOCWEATH SRCFILE(QRPGLESRC) DBGVIEW(*LIST)

     H DFTACTGRP(*NO) BNDDIR('HTTPAPI':'QC2LE')

      *
      * This controls the location of the message file
      * where weather information is stored.
      *
     D MSGFILE         C                   'QGPL/WEATHER'

      /copy httpapi_h
      /copy ifsio_h

      * This works like *ENTRY PLIST
     D LOCWEATH        PR                  ExtPgm('LOCWEATH')
     D   ZipCode                      5a   const
     D LOCWEATH        PI
     D   ZipCode                      5a   const


     D getAttr         PR           100a   varying
     D   name                       100a   varying const
     D   attr                          *   dim(32767)
     D                                     const options(*varsize)
     D crtmsgf         PR

     D NOAA_LatLonZipCode...
     D                 PR
     D   ZipCode                      5a   const
     D   lat                          8f
     D   lon                          8f
     D NOAA_NDFDgenByDay...
     D                 PR
     D   lat                          8f   value
     D   lon                          8f   value
     D   start                         d   datfmt(*iso) const

     D startTag        PR
     D   unused                        *   value
     D   depth                       10I 0 value
     D   ns                        1024A   varying const
     D   name                      1024A   varying const
     D   path                     24576A   varying const
     D   attrs                         *   dim(32767)
     D                                     const options(*varsize)
     D endTag          PR
     D   save                     65535a   varying
     D   depth                       10I 0 value
     D   ns                        1024A   varying const
     D   name                      1024A   varying const
     D   path                     24576A   varying const
     D   value                    65535A   varying const
     D   attrs                         *   dim(32767)
     D                                     const options(*varsize)
     D getTemp         PR             3p 0
     D   val                          3a   varying const
     D getDow          PR             9a   varying
     D   date                          d   const datfmt(*iso)

     D summ            s             40a   varying
     D temptype        s             10a   varying
     D tempunit        s             10a   varying
     D dayno           s             10i 0 inz(0)

     D weather         ds                  qualified
     D                                     dim(5) inz
     D   start                         d   datfmt(*iso)
     D   low                          3p 0
     D   high                         3p 0
     D   summary                     40a   varying

     D Msg             s             52a
     D lat             s              8f
     D lon             s              8f

      /free

       if ( %parms < 1 );
          http_comp( 'Please pass a zip code CALL LOCWEATH +
                      PARM(''53207'')');
          return;
       endif;

       CrtMsgF();

       http_debug(*ON:'/tmp/weather_debug.txt');

       http_XmlStripCRLF(*ON);
       http_XmlNs(*ON);

       monitor;
          NOAA_LatLonZipCode( ZipCode: lat: lon );
          NOAA_NDFDgenByDay( lat: lon: %date() );
       on-error;
          dump(A);
       endmon;

       *inlr = *on;
      /end-free


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * NOAA_LatLonZipCode():  Get the latitude and longitude of
      *                        a particular USA zip code.
      *
      *    ZipCode = (input) zip code to retrieve
      *        lat = (output) latitude
      *        lon = (output) longitude
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P NOAA_LatLonZipCode...
     P                 B
     D NOAA_LatLonZipCode...
     D                 PI
     D   ZipCode                      5a   const
     D   lat                          8f
     D   lon                          8f

     D SOAP            s           1024A   varying
     D LatLon          s          65535A   varying
     D LatLonList      s          65535A   varying
     D rc              s             10I 0
     D x               s             10I 0
     D pos             s             10I 0
     D errnum          s             10i 0

      /free
       SOAP =
          '<soapenv:Envelope +
              xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" +
              xmlns:xsd="http://www.w3.org/2001/XMLSchema" +
              xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" +
              xmlns:ndf="http://graphical.weather.gov/xml/+
                         DWMLgen/wsdl/ndfdXML.wsdl">+
           <soapenv:Header/>+
           <soapenv:Body>+
             <ndf:LatLonListZipCode +
                 soapenv:encodingStyle="http://schemas.xmlsoap.org/+
                                      soap/encoding/">+
               <zipCodeList xsi:type="xsd:string">' + ZipCode +
              '</zipCodeList>+
             </ndf:LatLonListZipCode>+
           </soapenv:Body>+
          </soapenv:Envelope>';

       // -------------------------------------------------------
       //   Ask the NOAA web service for the latitude/longitude
       //     - sometimes the NOAA returns a bad response,
       //       so try up to 5 times for a valid answer.
       //     - save the XML response in the LatLon variable.
       // -------------------------------------------------------

       for x = 1 to 5;
           rc = http_url_post_xml( 'http://graphical.weather.gov/xml/+
                                    SOAP_server/ndfdXMLserver.php'
                                 : %addr(SOAP) + 2
                                 : %len(SOAP)
                                 : *NULL
                                 : %paddr(endTag)
                                 : %addr(LatLon)
                                 : HTTP_TIMEOUT
                                 : HTTP_USERAGENT
                                 : 'text/xml'
                                 : '"http://graphical.weather.gov/+
                                     xml/DWMLgen/wsdl/ndfdXML.wsdl#+
                                     LatLonListZipCode"');
           if (rc = 1);
              leave;
           endif;

           http_error(errnum);
           if (errnum<>HTTP_XMLERR or x=5);
              http_crash();
              return;
           endif;

       endfor;


       // -------------------------------------------------------
       //   The response from the lat/lon lookup is an XML
       //   document.  Parse it to get the latitude and longitude
       //   (in the latLonList variable)
       // -------------------------------------------------------

       rc = http_parse_xml_string( %addr(latLon) + 2
                                 : %len(latLon)
                                 : 0
                                 : *NULL
                                 : %paddr(endTag)
                                 : %addr(latLonList) );
       if (rc < 0);
          http_crash();
          return;
       endif;


       // -------------------------------------------------------
       //   The latitude and longitude should now be in the
       //   latLonList variable, separated by a comma.  Split
       //   out the lat & lon from that list.
       // -------------------------------------------------------

       pos = %scan(',': latLonList);
       if (pos<2 or pos=%len(latLonList));
          lat = 0;
          lon = 0;
       else;
          lat = %float(%subst(latLonList:1:pos-1));
          lon = %float(%subst(latLonList:pos+1));
       endif;

      /end-free
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * NOAA_NDFDgetByDay():  Generate 5 day weather forecasat for a
      *                       given latitude & longitude
      *
      *     lat = (input) latitude
      *     lon = (input) longitude
      *   start = (input) date to start with
      *
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P NOAA_NDFDgenByDay...
     P                 B
     D NOAA_NDFDgenByDay...
     D                 PI
     D   lat                          8f   value
     D   lon                          8f   value
     D   start                         d   datfmt(*iso) const

     D QCMDEXC         PR                  ExtPgm('QCMDEXC')
     D   cmd                       3000a   const options(*varsize)
     D   len                         15p 5 const
     D   igc                          3a   const options(*nopass)

     D SOAP            s          65535A   varying
     D RESP            s          65535A   varying
     D rc              s             10I 0
     D cmd             s           2000a   varying
     D x               s              3p 0
     D errnum          s             10i 0
      /free
       SOAP=
         '<soapenv:Envelope +
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" +
            xmlns:xsd="http://www.w3.org/2001/XMLSchema" +
            xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" +
            xmlns:ndf="http://graphical.weather.gov/xml/+
              DWMLgen/wsdl/ndfdXML.wsdl">+
          <soapenv:Header/>+
          <soapenv:Body>+
            <ndf:NDFDgenByDay +
              soapenv:encodingStyle="http://schemas.xmlsoap.org/+
                soap/encoding/">+
              <latitude xsi:type="xsd:decimal">'
                 + %char(%dec(lat:9:6)) +
              '</latitude>+
              <longitude xsi:type="xsd:decimal">'
                 + %char(%dec(lon:9:6)) +
              '</longitude>+
              <startDate xsi:type="xsd:date">'
                 + %char(start:*ISO) +
              '</startDate>+
              <numDays xsi:type="xsd:integer">5</numDays>+
              <Unit xsi:type="xsd:string">e</Unit>+
              <format xsi:type="xsd:string">24 hourly</format>+
            </ndf:NDFDgenByDay>+
         </soapenv:Body>+
         </soapenv:Envelope>';

       // -------------------------------------------------------
       //  Ask the NOAA for the 5 day weather outlook.
       //  store the response in the 'RESP' variable
       //
       //  - Occasionally, the NOAA returns an invalid
       //     response.  Try up to 5 times to get a valid one.
       // -------------------------------------------------------

       for x = 1 to 5;

           rc = http_url_post_xml( 'http://graphical.weather.gov/xml/+
                                    SOAP_server/ndfdXMLserver.php'
                                 : %addr(SOAP) + 2
                                 : %len(SOAP)
                                 : *NULL
                                 : %paddr(endTag)
                                 : %addr(RESP)
                                 : HTTP_TIMEOUT
                                 : HTTP_USERAGENT
                                 : 'text/xml'
                                 : '"http://graphical.weather.gov/+
                                    xml/DWMLgen/wsdl/ndfdXML.wsdl#+
                                    NDFDgenByDay"');
           if (rc = 1);
               leave;
           endif;

           http_error(errnum);
           if (errnum<>HTTP_XMLERR or x=5);
              http_crash();
              return;
           endif;
       endfor;

       // -------------------------------------------------------
       //  Parse the XML document in the RESP variable to get
       //  the weather information.
       // -------------------------------------------------------

       reset weather;
       reset summ;
       reset temptype;
       reset tempunit;
       reset dayno;

       rc = http_parse_xml_string( %addr(RESP) + 2
                                 : %len(RESP)
                                 : 0
                                 : %paddr(startTag)
                                 : %paddr(endTag)
                                 : *NULL );
       if (rc < 0);
          http_crash();
          return;
       endif;

       // -------------------------------------------------------
       //  Load the weather information into message descriptions
       //  in the WEATHER message file
       // -------------------------------------------------------

       for x = 1 to %elem(Weather);
          cmd = 'CHGMSGD MSGID(DAY0' + %editc(x:'X') + ') +
                         MSGF(' + MSGFILE + ') +
                         MSG(''' + getDow(weather(x).start) + ''')';
          QCMDEXC(cmd: %len(cmd));
          cmd = 'CHGMSGD MSGID(DAT0' + %editc(x:'X') + ') +
                         MSGF(' + MSGFILE + ') +
                         MSG(''' + %char(weather(x).start:*MDY/) + ''')';
          QCMDEXC(cmd: %len(cmd));
          cmd = 'CHGMSGD MSGID(HIG0' + %editc(x:'X') + ') +
                         MSGF(' + MSGFILE + ') +
                         MSG(''' + %editc(weather(x).high:'3') + ''')';
          QCMDEXC(cmd: %len(cmd));
          cmd = 'CHGMSGD MSGID(LOW0' + %editc(x:'X') + ') +
                         MSGF(' + MSGFILE + ') +
                         MSG(''' + %editc(weather(x).low:'3') + ''')';
          QCMDEXC(cmd: %len(cmd));
          cmd = 'CHGMSGD MSGID(SUM0' + %editc(x:'X') + ') +
                         MSGF(' + MSGFILE + ') +
                         MSG(''' + weather(x).summary + ''')';
          QCMDEXC(cmd: %len(cmd));
       endfor;


      /end-free
     P                 E



      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * startTag(): This is called by HTTPAPI for each starting
      *             XML tag that's received over the network
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P startTag        B
     D startTag        PI
     D   unused                        *   value
     D   depth                       10I 0 value
     D   ns                        1024A   varying const
     D   name                      1024A   varying const
     D   path                     24576A   varying const
     D   attrs                         *   dim(32767)
     D                                     const options(*varsize)
      /free
          if ( path='/dwml/data' and name='time-layout' );
              summ = getAttr('summarization': attrs);
              dayno = 0;
          endif;

          if ( path='/dwml/data/parameters' and name='temperature' );
              temptype = getAttr('type': attrs);
              tempunit = getAttr('units': attrs);
              dayno = 0;
          endif;

          if ( path='/dwml/data/parameters' and name='weather' );
              dayno = 0;
          endif;
      /end-free
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * endTag(): This is called by HTTPAPI for each ending XML tag
      *           that's received over the network
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P endTag          B
     D endTag          PI
     D   save                     65535a   varying
     D   depth                       10I 0 value
     D   ns                        1024A   varying const
     D   name                      1024A   varying const
     D   path                     24576A   varying const
     D   value                    65535A   varying const
     D   attrs                         *   dim(32767)
     D                                     const options(*varsize)

      /free
          if path='/Envelope/Body/LatLonListZipCodeResponse'
             and name = 'listLatLonOut';
              save = value;
          endif;
          if path='/dwml' and name='latLonList';
              save = value;
          endif;
          if path='/Envelope/Body/NDFDgenByDayResponse'
             and name = 'dwmlByDayOut';
              save = value;
          endif;

          if ( path='/dwml/data/time-layout'
               and name='start-valid-time'
               and summ='24hourly'
               and dayno<%elem(weather) );
                 dayno = dayno + 1;
                 weather(dayno).start = %date(%subst(value:1:10):*ISO);
          endif;

          if ( path='/dwml/data/parameters/temperature'
              and tempunit='Fahrenheit'
              and name = 'value'
              and dayno<%elem(weather) );

              select;
              when  temptype='minimum';
                 dayno = dayno + 1;
                 weather(dayno).low = getTemp(value);
              when  temptype='maximum';
                 dayno = dayno + 1;
                 weather(dayno).high = getTemp(value);
              endsl;

          endif;

          if ( path='/dwml/data/parameters/weather'
              and name = 'weather-conditions'
              and dayno < %elem(weather) );
              dayno = dayno + 1;
              weather(dayno).summary = getAttr('weather-summary': attrs);
              if (weather(dayno).summary=''
                  or weather(dayno).summary=*blanks );
                  weather(dayno).summary='-';
              endif;
          endif;

      /end-free
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * getAttr():  Search for a particular attribute in an
      *             element's attribute list
      *
      *    name = (input) Attribute to search for
      *    attr = (input) list of attributes to search
      *
      * returns the value of the attribute, or '' if not found.
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P getAttr         B
     D getAttr         PI           100a   varying
     D   name                       100a   varying const
     D   attr                          *   dim(32767)
     D                                     const options(*varsize)
     D x               s             10i 0 inz(1)
      /free

          dow attr(x) <> *NULL;
             if %str(attr(x)) = name;
                return %str(attr(x+1));
             endif;
             x = x + 2;
          enddo;

          return '';
      /end-free
     P                 E



      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * getTemp():  Convert a character string containing a
      *             temperature to a numeric value
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P getTemp         B
     D getTemp         PI             3p 0
     D   val                          3a   varying const
     D rc              s              3p 0
      /free
         monitor;
            rc = %dec(val:3:0);
         on-error;
            rc = 0;
         endmon;
         return rc;
      /end-free
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * getDow():  Calculate the day of the week for a given date
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P getDow          B
     D getDow          PI             9a   varying
     D   date                          d   const datfmt(*iso)
     D SUNDAY          c                   d'2006-01-01'
     D Count           s             10i 0
     D Day             s             10i 0
      /free
          count = %diff(date:SUNDAY:*DAYS);
          day = %rem(count:7);
          select;
          when day = 0;
             return 'Sunday';
          when day = 1;
             return 'Monday';
          when day = 2;
             return 'Tuesday';
          when day = 3;
             return 'Wednesday';
          when day = 4;
             return 'Thursday';
          when day = 5;
             return 'Friday';
          other;
             return 'Saturday';
          endsl;
      /end-free
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * getDow():  Calculate the day of the week for a given date
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P crtmsgf         B
     D crtmsgf         PI

     D QCMDEXC         PR                  ExtPgm('QCMDEXC')
     D   cmd                       3000a   const options(*varsize)
     D   len                         15p 5 const
     D   igc                          3a   const options(*nopass)

     D cmd             s           3000a   varying
     D found           s              1n   inz(*OFF)
     D x               s              3s 0

      /free

          monitor;
            cmd = 'CHKOBJ OBJ(' + MSGFILE + ') OBJTYPE(*MSGF)';
            QCMDEXC(cmd:%len(cmd));
            Found = *ON;
          on-error;
            cmd = 'CRTMSGF MSGF(' + MSGFILE + ') +
                           TEXT(''Five day weather outlook'')';
            QCMDEXC(cmd:%len(cmd));
            Found = *OFF;
          endmon;

          if (not Found);
             for x = 1 to 5;
                cmd = 'ADDMSGD MSGID(DAY0' + %editc(x:'X') + ') +
                               MSGF(' + MSGFILE + ') +
                               MSG(''-'')';
                QCMDEXC(cmd:%len(cmd));
                cmd = 'ADDMSGD MSGID(DAT0' + %editc(x:'X') + ') +
                               MSGF(' + MSGFILE + ') +
                               MSG(''-'')';
                QCMDEXC(cmd:%len(cmd));
                cmd = 'ADDMSGD MSGID(HIG0' + %editc(x:'X') + ') +
                               MSGF(' + MSGFILE + ') +
                               MSG(''-'')';
                QCMDEXC(cmd:%len(cmd));
                cmd = 'ADDMSGD MSGID(LOW0' + %editc(x:'X') + ') +
                               MSGF(' + MSGFILE + ') +
                               MSG(''-'')';
                QCMDEXC(cmd:%len(cmd));
                cmd = 'ADDMSGD MSGID(SUM0' + %editc(x:'X') + ') +
                               MSGF(' + MSGFILE + ') +
                               MSG(''-'')';
                QCMDEXC(cmd:%len(cmd));
             endfor;
          endif;

          return;
      /end-free
     P                 E
